########################################################################################################################
# Database differences between two databases
########################################################################################################################

require_relative '../model/database'
require_relative 'diff'
require_relative 'index_comparator'
require_relative 'table_comparator'
require_relative 'trigger_comparator'
require_relative 'view_comparator'

module Dbsketch
	module Comparison

		class DatabaseDiff < Diff
			def initialize(old_database, new_database, indexes, operations, tables, triggers, views)
				super old_database, new_database
				### Preconditions
				raise ArgumentError, "old_database is not a Dbsketch::Model::Database" unless old_database.is_a? Dbsketch::Model::Database
				raise ArgumentError, "new_database is not a Dbsketch::Model::Database" unless new_database.is_a? Dbsketch::Model::Database
				###
				@indexes = indexes
				@operations = operations
				@tables = tables
				@triggers = triggers
				@views = views
			end
			attr_reader :indexes, :operations, :tables, :triggers, :views
		end

		class DatabaseComparator
			def initialize options: {}
				@index_comparator = IndexComparator.new
				@function_comparator = FunctionComparator.new
				@procedure_comparator = ProcedureComparator.new
				@table_comparator = TableComparator.new :options => options
				@trigger_comparator = TriggerComparator.new
				@view_comparator = ViewComparator.new
			end

			def are_equivalent? old_database, new_database
				### Preconditions
				raise ArgumentError, "old_database is not a Dbsketch::Model::Database" unless old_database.is_a? Dbsketch::Model::Database
				raise ArgumentError, "new_database is not a Dbsketch::Model::Database" unless new_database.is_a? Dbsketch::Model::Database
				###
				indexes(old_database, new_database).empty? and operations(old_database, new_database).empty? and tables(old_database, new_database).empty? and triggers(old_database, new_database).empty? and views(old_database, new_database).empty?
			end

			# Returns a DatabaseDiff if tables are different, nil otherwise
			def compare old_database, new_database
				### Preconditions
				raise ArgumentError, "old_database is not a Dbsketch::Model::Database" unless old_database.is_a? Dbsketch::Model::Database
				raise ArgumentError, "new_database is not a Dbsketch::Model::Database" unless new_database.is_a? Dbsketch::Model::Database
				###
				DatabaseDiff.new(
					old_database, new_database, indexes(old_database, new_database), operations(old_database, new_database), tables(old_database, new_database), triggers(old_database, new_database), views(old_database, new_database)
				) if not are_equivalent? old_database, new_database
			end

			private

			def find_diffs old_database, new_database, comparator, collection
				### Preconditions
				raise ArgumentError, "collection is not a Symbol" unless collection.is_a? Symbol
				raise ArgumentError, "collection is not a method of Dbsketch::Model::Database" unless old_database.respond_to? collection
				###
				diffs = old_database.method(collection).call.select { |old_item| not new_database.has_item? old_item.name }.map do |old_item|
					comparator.compare(old_item, nil)
				end
				diffs << new_database.method(collection).call.select { |new_item| not old_database.has_item? new_item.name }.map do |new_item|
					comparator.compare(nil, new_item)
				end
				diffs << old_database.method(collection).call.select { |old_item| new_database.has_item? old_item.name }.map do |old_item|
					comparator.compare(old_item, new_database[old_item.name])
				end
				diffs.flatten.compact
			end

			def indexes old_database, new_database
				find_diffs old_database, new_database, @index_comparator, :indexes
			end

			def operations old_database, new_database
				diffs = old_database.operations.select { |old_op| not new_database.has_item? old_op.name }.map do |old_op|
					(old_op.is_a? Dbsketch::Model::Function) ? @function_comparator.compare(old_op, nil) : @procedure_comparator.compare(old_op, nil)
				end
				diffs << new_database.operations.select { |new_op| not old_database.has_item? new_op.name }.map do |new_op|
					(new_op.is_a? Dbsketch::Model::Function) ? @function_comparator.compare(nil, new_op) : @procedure_comparator.compare(nil, new_op)
				end
				diffs << old_database.operations.select { |old_op| new_database.has_item? old_op.name }.map do |old_op|
					if old_op.class == new_database[old_op.name].class
						(old_op.is_a? Dbsketch::Model::Function) ? @function_comparator.compare(old_op, new_database[old_op.name]) : @procedure_comparator.compare(old_op, new_database[old_op.name])
					else
						Diff.new old_op, new_database[old_op.name]
					end
				end
				diffs.flatten.compact
			end

			def tables old_database, new_database
				find_diffs old_database, new_database, @table_comparator, :tables
			end

			def triggers old_database, new_database
				find_diffs old_database, new_database, @trigger_comparator, :triggers
			end

			def views old_database, new_database
				find_diffs old_database, new_database, @view_comparator, :views
			end

		end

	end
end
